

## This ensures that the next import requests will be in the correct directory
## i.e., we get the path to this file and we move to it
import inspect,os
working_directory = os.getcwd()
source_directory = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
os.chdir(source_directory)

load('./linear_ordering.py')
load('./graph-io.py')
load('./io-tools.py')

## Now that the code is loaded, we go back to the working directory
os.chdir(working_directory)







latex_name = {
    'pw': {
        'AD1': "\\ilp{PW}{alldiff:bad}",
        'PD1': "\\ilp{PW}{alldiff}",
        'SS': "\\ilp{PW}{alldiff:v3}",
        'RSoPi10': "\\cite{SoPi10}",
        'ADuarte12': "\\cite{Duarte2012}",
        'BBNNPR': "\\cite{DBLP:conf/gd/BiedlBNNPR13}",
        'PD2': "PD2"
        },
    'cw': {
        'AD1': "\\ilp{CW}{alldiff:bad}",
        'PD1': "\\ilp{CW}{alldiff}",
        'SD1': "\\ilp{CW}{alldiff:v3}",
        'ACK': "\\ilp{CW'}{alldiff:bad}",
        'PCK': "\\ilp{CW'}{alldiff}",
        'ALPSY05': "\\cite{DBLP:conf/cata/LuttamaguziPSY05}",
        'SLL+14': "\\cite{DBLP:series/sci/LocesCHBPRBV14}"
        },
    'bw': {
        'AD1': "\\ilp{BW}{alldiff:bad}",
        'PD1': "\\ilp{BW}{alldiff}",
        'SD1': "\\ilp{BW}{alldiff:v3}",
        'APinana04': "\\cite{Pinana04}",
        'PPinana04': "\\cite{Pinana04}+F\\ref{alldiff}",
        'SPinana04': "\\cite{Pinana04}+F\\ref{alldiff:v3}",
        'ALPSY05': "\\cite{DBLP:conf/cata/LuttamaguziPSY05}",
        'BBNNPR': "\\cite{DBLP:conf/gd/BiedlBNNPR13}"
        },
    'sc': {
        'AD1': "\\ilp{SC}{alldiff:bad}",
        'PD1': "\\ilp{SC}{alldiff}",
        'SD1': "\\ilp{SC}{alldiff:v3}",
        'AD2': "\\ilp{PR}{alldiff:bad}",
        'PD2': "\\ilp{PR}{alldiff:bad}",
        'SD2': "\\ilp{PR}{alldiff:v3}",
        'APinana04': "\\cite{Pinana04}",
        'PPinana04': "\\cite{Pinana04}+F\\ref{alldiff}",
        'SPinana04': "\\cite{Pinana04}+F\\ref{alldiff:v3}"
        },
    'ola': {
        'AD1': "\\ilp{OLA}{alldiff:bad}",
        'PD1': "\\ilp{OLA}{alldiff}",
        'SD1': "\\ilp{OLA}{alldiff:v3}",
        'AD2': "OLA4",
        'PD2': "OLA5",
        'SD2': "OLA6",
        'ACK': "OLA7",
        'PCK': "OLA8",
        'APinana04': "\\cite{Pinana04}",
        'PPinana04': "\\cite{Pinana04}+F\\ref{alldiff}",
        'SPinana04': "\\cite{Pinana04}+F\\ref{alldiff:v3}"
        }
    }










################################################################################
# 
################################################################################

@parallel
def myrun(G, prob, model, solver=None, timelimit=1800, verbose=0, check=True, logfile=None, nthreads=None):
    """
    """
    GL = GraphLayout(G, solver=solver, timelimit=timelimit, verbose=verbose,
                     check=check, logfile=logfile, nthreads=nthreads)

    if prob == 'pw':
        if model=='BBNNPR':
            obj = GL.pathwidth_BBNNPR()
        else:
            obj,order = GL.pathwidth(model)
    elif prob == 'cw':
        obj,order = GL.cutwidth(model)
    elif prob == 'bw':
        if model=='BBNNPR':
            obj = GL.bandwidth_BBNNPR()
        else:
            obj,order = GL.bandwidth(model)
    elif prob == 'sc':
        obj,order = GL.sumcut(model)
    elif prob == 'ola':
        obj,order = GL.optimal_linear_arrangement(model)

    if GL.time < timelimit*.99:
        gap, bestobj, isopt = 0.0, float(obj), True
    else:
        gap     = GL.get_relative_mip_gap()
        bestobj = GL.get_best_objective_value()
        isopt   = gap<1.0*10**(-4)
    res = {'bestint':obj, 'bestobj':bestobj, 'gap':gap, 'opt':isopt, 'time': round(GL.time,4),
           'order':['*']*GL._N if model=='BBNNPR' else order}
    print("{} / {}: {} ({}, {})\t {}".format(prob,model,G.name(),G.order(),G.size(),res))
    return res


def experiment_one_graph(G, solver=None, timelimit=None, verbose=0, check=True, logfile=None, nthreads=None):
    """
    """
    from itertools import product
    #GL = GraphLayout(G, solver=solver, timelimit=timelimit, verbose=verbose,
    #                 check=check, logfile=logfile, nthreads=nthreads)

    res = {'n': G.order(), 'm':G.size(), 'pw':{}, 'cw':{}, 'bw':{}, 'sc':{}, 'ola':{}}

    my_models = {}
    
    # PATHWIDTH
    #my_models = ['AD1']+list(a+b for a,b in product(['P','Q','R'], ['D1', 'D2', 'D3', 'SoPi10']))+['SS']
    my_models['pw'] = ['AD1','PD1','PD2','RSoPi10','SS', 'ADuarte12', 'BBNNPR']

    # CUTWIDTH
    #my_models = ['ALPSY05']+list(a+b for a,b in product(['A','P','Q','R'], ['D1', 'CK']))+['SD1', 'SLL+14']
    my_models['cw'] = ['AD1','PD1','SD1','ACK','PCK','ALPSY05','SLL+14']

    # BANDWIDTH
    my_models['bw'] = list(a+b for a,b in product(['A','P','S'], ['D1','Pinana04'])) + ['ALPSY05', 'BBNNPR']

    # SUMCUT
    my_models['sc'] = list(a+b for a,b in product(['A','P','S'], ['D1','D2']))

    # OPTIMAL LINEAR ARRANGEMENT
    my_models['ola'] = list(a+b for a,b in product(['A','P','S'], ['D1','D2'])) + ['ACK', 'PCK']

    if timelimit is None:
        timelimit = 10**6
    """
    for prob in sorted(my_models):
        for model in sorted(my_models[prob]):
            res[prob][model] = myrun(G, prob, model, solver=solver, timelimit=timelimit, verbose=verbose,
                                        check=check, logfile=logfile, nthreads=nthreads)
    """
    Q = list(myrun([ (G, prob, model, solver, timelimit, verbose, check, logfile, nthreads)
                      for prob in sorted(my_models) for model in sorted(my_models[prob]) ] ) )
    for p,sol in Q:
        prob = p[0][1]
        model = p[0][2]
        res[prob][model] = sol
        
    return res

            
################################################################################
# 
################################################################################

def experiment_many_graphs(source_dir, dest_dir, timelimit, solver=None, nthreads=None, nmax=None):
    """
    """
    source_dir += '' if source_dir[-1]=='/' else '/'
    dest_dir += '' if dest_dir[-1]=='/' else '/'
    nmax = 10**5 if nmax is None else nmax
    GIO = GraphIO()
    listdir = sorted(os.listdir(source_dir))
    for name in listdir:
        G = GIO.read_edgelist(source_dir+name)
        if not G.order() or G.order() > nmax:
            continue
        G.name(name)
        print('='*80+'\n= '+name+'\n'+'='*80)
        print('n = {}\nm = {}\n'.format(G.order(),G.size()))
        res = experiment_one_graph(G, solver=solver, timelimit=timelimit, nthreads=nthreads)
        res['name'] = name
        if name.endswith('edgelist'):
            write_dictionary_to_file(res, filename=dest_dir+name.replace('edgelist','json'))
        else:
            write_dictionary_to_file(res, filename=dest_dir+name+'.json')



def run_extra_stuff(source_dir, dest_dir, prob='bw', model='APinana04', timelimit=300, solver=None, nthreads=None, nmax=None):
    """
    """
    source_dir += '' if source_dir[-1]=='/' else '/'
    nmax = 10**5 if nmax is None else nmax
    GIO = GraphIO()
    listdir = sorted(os.listdir(source_dir))
    list_of_stuff = []
    for name in listdir:
        G = GIO.read_edgelist(source_dir+name)
        if G.order()==0 or G.order()>nmax:
            continue
        G.name(name)
        list_of_stuff.append( (G, prob, model, solver, timelimit, False, True, None, nthreads) )

    Q = list(myrun(list_of_stuff))

    res = {}    
    for p,sol in Q:
        G = p[0][0]
        res[G.name()] = {prob:{model:sol}}
    
    write_dictionary_to_file(res, filename=dest_dir+prob+'-'+model+'.json')


def run_extra_stuff_long(source_dir, dest_dir, prob='bw', model='APinana04', timelimit=300, solver=None, nthreads=None, nmax=None, step=100, skip_blocks=0, max_unsolved=400):
    """
    """
    source_dir += '' if source_dir[-1]=='/' else '/'
    nmax = 10**5 if nmax is None else nmax
    GIO = GraphIO()
    listdir = sorted(os.listdir(source_dir))
    L = []
    for name in listdir:
        G = GIO.read_edgelist(source_dir+name)
        if G.order()==0 or G.order()>nmax:
            continue
        G.name(name)
        #L.append( (G.order(),G.size(),G) )
        L.append( (G.order()+G.size(),G.order(),G.size(),G) )

    res = {G.name():{"n":n,"m":m,"name":G.name()} for _,n,m,G in L}
    write_dictionary_to_file(res, filename=dest_dir+'zz_list_of_graphs.json')


    #list_of_stuff = list( (G, prob, model, solver, timelimit, False, True, None, nthreads) for _,_,G in sorted(L) )
    list_of_stuff = list( (G, prob, model, solver, timelimit, False, True, None, nthreads) for _,_,_,G in sorted(L) )
    nb_stuff = len(list_of_stuff)
    blocks_of_stuff = list( list_of_stuff[step*i: step*(i+1) if step*(i+1)<=nb_stuff else nb_stuff] for i in range(ceil(nb_stuff/step)) )
    nb_blocks = len(blocks_of_stuff)
    digits = len(str(nb_blocks))
    int_to_str = lambda _i:'0'*(digits-len(str(_i)))+str(_i)

    cpt_unsolved = 0
    for i,list_of_stuff in enumerate(blocks_of_stuff):
        if i<skip_blocks:
            print("skip block {}".format(i+1))
            continue
        print("\n>>>> {}/{} <<<<\n".format(i+1,nb_blocks))
        Q = list(myrun(list_of_stuff))

        res = {}
        for p,sol in Q:
            G = p[0][0]
            res[G.name()] = {prob:{model:sol}}
            if sol['opt']:
                cpt_unsolved = 0
            else:
                cpt_unsolved += 1

        write_dictionary_to_file(res, filename=dest_dir+prob+'-'+model+'-'+int_to_str(i+1)+'.json')

        if cpt_unsolved>=max_unsolved:
            print("{} consecutive unsolved instances => STOP".format(max_unsolved))
            break





    
################################################################################
# 
################################################################################

def merge_dictionary(source_dir, dest_file):
    """
    """
    listdir = os.listdir(source_dir)

    if 'zz_list_of_graphs.json' in listdir:
        data = read_dictionary_from_file(source_dir+'/zz_list_of_graphs.json')
        listdir.remove('zz_list_of_graphs.json')
    else:
        data = {}
    for filename in listdir:
        tmp = read_dictionary_from_file(source_dir+'/'+filename)
        if "name" in tmp:
            name = tmp["name"]
            if name in data:
                for z in tmp:
                    if z=='name':
                        continue
                    if isinstance(tmp[z], dict):
                        if not z in data[name]:
                            data[name][z] = {}
                        for model in tmp[z]:
                            data[name][z][model] = tmp[z][model]
                    else:
                        data[name][z] = tmp[z]
            else:
                data[name] = tmp
        else:
            l = filename.replace('.json','').split('-')
            prob,model = l[:2]
            for name in tmp:
                if name in data:
                    if not prob in data[name]:
                        data[name][prob] = tmp[name][prob]
                    else:
                        data[name][prob][model] = tmp[name][prob][model]
                else:
                    data[name] = tmp[name]

    write_dictionary_to_file(data, filename=dest_file)




################################################################################
#
################################################################################

def stat_prob(filename, prob):
    """
    """
    data = read_dictionary_from_file(filename)

    cpt = {}
    for name in data:
        for model in data[name][prob]:
            if not model in cpt:
                cpt[model] = {'opt':0, 'gap':0.0, 'time':0.0}
            cpt[model]['opt'] += data[name][prob][model]['opt']
            cpt[model]['gap'] += data[name][prob][model]['gap']
            cpt[model]['time'] += data[name][prob][model]['time']

    n = len(data)
    for model in cpt:
        cpt[model]['gap'] = round(cpt[model]['gap']/n,4)
        cpt[model]['time'] = round(cpt[model]['time']/n,4)

    return cpt



def gen_small_latex_tab(filename, prob, order=None):
    """
    """
    data = stat_prob(filename, prob)
    print("Available models: {}".format(data.keys()))
    order = sorted(data.keys()) if order is None else order
    st  = "\\begin{tabular}{l|"+'d'*len(order)+"}\n\\hline"
    st += "\t"
    for model in order:
        st += "& \\multicolumn{1}{c}{"+latex_name[prob][model]+"} "
    st += "\\\\\\hline\n"
    st += "\\# opt\t"
    for model in order:
        st += "& {} ".format(data[model]['opt'])
    st += "\\\\\\hline\n"
    st += "\\% gap\t"
    for model in order:
        st += "& {} ".format(round(100*data[model]['gap'],1))
    st += "\\\\\\hline\n"
    st += "CPU time\t"
    for model in order:
        st += "& {} ".format(round(data[model]['time'],1))
    st += "\\\\\\hline\n\\end{tabular}\n"
    print(st)
    


