"""
This file gathers useful methods for read/write operations on files.


"""
import json


def write_dictionary_to_file(dico, filename=None, comment=None, standard=False, indent=2, sort_keys=True, verbose=False):
    """
    """
    char_iterable_open = {list:'[', dict:'{', tuple:'('}
    char_iterable_close = {list:']', dict:'}', tuple:')'}
    
    def good_str(elt, indent_base, indent):
        """
        If elt is string, return ``"elt"``
        else return str(elt)
        """
        if isinstance(elt, str):
            if not '"' in elt and not elt[0]=="'":
                st = '"'+str(elt)+'"'
            elif not "'" in elt and not elt[0]=='"':
                st = "'"+str(elt)+"'"
            else:
                st = str(elt)

        elif isinstance(elt, list):
            if elt and (isinstance(elt[0], list) or (isinstance(elt[0], tuple) and len(elt[0])>2)):
                st = char_iterable_open[type(elt[0])]+'\n'
                for i,v in enumerate(elt):
                    st += ' '*(indent_base+indent)+good_str(v, indent_base+indent, indent)
                    st += '\n' if i==len(elt)-1 else ',\n'
                st += ' '*indent_base+char_iterable_close[type(elt[0])]
            else:
                st = str(elt)

        elif isinstance(elt, dict):
            if any(isinstance(elt[key], dict) for key in elt):
                st = '{\n'
                for i,v in enumerate(sorted(elt)):
                    st += ' '*(indent_base+indent)+good_str(v, indent_base+indent, indent)
                    st += ': '+good_str(elt[v], indent_base+indent, indent)
                    st += '\n' if i==len(elt)-1 else ',\n'
                st += ' '*indent_base+'}'
            else:
                st = str(elt)

        else:
            st = str(elt)

        return st


    #try:
    if True:
        with open(filename, 'w') as f:
            if verbose:
                print("Succeed to open file", filename)
            if comment is not None:
                f.write("# "+str(comment)+"\n#\n")

            if standard:
                f.write(json.dumps(dico, indent=indent, sort_keys=sort_keys))
            else:
                try: # test if we have a dict of dict
                    keys = next(dico.itervalues()).keys()
                except:
                    raise ValueError("The input dictionary is not a dictionary of dictionaries.")

                # Write dictionary to file
                my_indent = ' '*indent
                f.write("{\n")
                my_order = sorted( dico.keys() ) if sort_keys else dico.keys()
                for elt in my_order:
                    if isinstance(dico[elt],dict):
                        f.write(my_indent+good_str(elt, indent, indent)+': {\n')
                        keys = sorted( dico[elt].keys() ) if sorted else dico[elt].keys()
                        for k in keys:
                            f.write(my_indent*2+good_str(k, 2*indent, indent)+': '+
                                    good_str(dico[elt][k], 2*indent, indent)+
                                    ('\n' if k==keys[-1] else ',\n'))
                        f.write(my_indent+('}\n' if elt==my_order[-1] else '},\n'))
                    else:
                        f.write(my_indent+good_str(elt, indent, indent)+': '+good_str(dico[elt], indent, indent)+',\n')
                f.write('}\n')

            f.close()

#    except:
#        raise ValueError('Unable to open file: '+filename)


def read_dictionary_from_file(filename=None):
    """
    Read data from file 'filename' and return the dictionary.
    """
    if not filename:
        raise ValueError("No file name given.")

    try:
        with open(filename, 'rb') as f:
            data = eval(f.read())
            f.close()
    except:
        raise ValueError('Unable to read data from file '+filename)

    if isinstance(data, dict):
        return data
    else:
        raise ValueError("No dictionary found in file "+filename)




#===============================================================================
#===== Data recorder ===========================================================
#===============================================================================

class DataRecorder:

    def __init__(self, verbose=False):
        """
        """
        self._number_of_entries = 0
        self._session = None
        self._data = {}
        self._verbose = verbose

    def __dealloc__(self):
        """
        """
        if self._data:
            del self._data

    def reset(self):
        """
        """
        self._number_of_entries = 0
        if self._data:
            del self._data
        self._data = {}
        self._verbose = verbose

    def __len__(self):
        return self._number_of_entries


    def __copy__(self):
        """
        return a copy of self.
        """
        other = DataRecorder(verbose=self._verbose)
        other._data = deepcopy(self._data)
        other._number_of_entries = self._number_of_entries


    def open_session(self, id=None):
        """
        Initialize a new entry to store data
        """
        if id is not None and not id in self._data:
            self._session = id
        else:
            # We search for an unused key in [0.._number_of_entries]
            for i in range(self._number_of_entries, -1, -1):
                if not i in self._data:
                    self._session = i
                    break

        self._data[self._session] = {}
        self._number_of_entries += 1


    def get_session_id(self):
        """
        Return the id of the current session
        """
        if self._session is None:
            raise ValueError("No open session.")
        return self._session

    def close_session(self):
        """
        """
        self._session = None

    def reset_session(self):
        """
        Erase data stored during the session
        """
        id = self.get_session_id()
        del self._data[id]
        self._data[id] = {}

    def delete_session(self, id):
        """
        Delete all entries of session id, and so reduces the total number of
        entries.
        """
        if id in self._data:
            del self._data[id]
        self._number_of_entries -= 1


    def session_iterator(self, sort_sessions=False):
        """
        """
        if sort_sessions:
            keys = sorted( k for k in self._data )
        else:
            keys = self._data
        return iter(keys)
        

    def session_iterator2(self, key, dfunk=None, funk_for_key=None):
        """
        Return an iterator of the sessions.
        The sessions are ordered according the values for key.

        dfunk is a dictionary of lambdas. Each lambda is a test. If all keys are
        in the current record (another test), then it applies all lambdas to
        dfunk[x](self._data[id]). They must be True.

        funk_for_key is a special lambda for parameter key. It is useful for
        instance if all names are like 'grafo1234.56.edgelist' and we want to
        order by increasing values of 1234.56.
        """
        funk_for_key = (lambda _x_:_x_) if funk_for_key is None else funk_for_key 
        if key is None:
            return iter( self._data )
        if dfunk:
            L = list()
            for id in self.session_iterator():
                if any( x not in self._data[id] for x in dfunk ):
                    continue
                if all( dfunk[x](self._data[id]) for x in dfunk if dfunk[x] is not None):
                    L.append( (funk_for_key(self._data[id][key]), id) )
            L = sorted( L )
        else:
            L = sorted( (funk_for_key(self._data[id][key]), id) for id in self.session_iterator() )
        return iter( id for _,id in L )


    def record_item(self, key, item):
        """
        """
        id = self.get_session_id()
        self._data[id][key] = item
            

    def record_items(self, bunch):
        """
        Record all items in bunch.

        bunch is an iterable
        """
        try:
            if isinstance(bunch, dict):
                for k in bunch:
                    self.record_item(k,bunch[k])
            elif isinstance(bunch, list) or isinstance(bunch, tuple):
                for k,v in bunch:
                    self.record_item(k,v)
            else:
                raise ValueError("Something goes wrong...")
        except:
            raise ValueError("Unrecognize entry format for "+str(bunch))


    def get_values(self, id):
        """
        """
        if id in self._data:
            return self._data[id]
        else:
            raise ValueError("Unknown entry id: "+str(id))


    def get_value(self, id, key):
        """
        """
        if id in self._data and key in self._data[id]:
            return self._data[id][key]
        else:
            raise ValueError("Unknown entry id (%s) or key (%s)." %(str(id), str(key)))


    def get_set_of_keys(self):
        """
        Return the set of all known keys
        """
        known_keys = set()
        for i in self._data:
            for x in self._data[i]:
                known_keys.add(x)
        return known_keys


    def store_to_file(self, filename=None, standard=False, comment=None):
        """
        Store recorded data to file 'filename'.
        """
        if self._number_of_entries==0:
            print("WARNING: no recorded data.\nabort")
            return
        elif filename is None:
            print("WARNING: no output file name given.\nabort")
            return
            
        write_dictionary_to_file(self._data, filename=filename, standard=standard,
                                 comment=comment, verbose=self._verbose)


    def read_from_file(self, filename=None, reset=True):
        """
        Read data from file 'filename'.

        If reset is 'True', former data are erased. Else, new data are added to
        the current one.
        """
        if reset:
            self.reset()

        # Read dictionary from file
        data = read_dictionary_from_file(filename)

        # We now update data
        for k in sorted(data.keys()):
            self.open_session(k)
            self.record_items(data[k])
            self.close_session()



    def get_values_for_key(self, key, req=None, dfunk=None, key_order=None, sort_values=False):
        """
        req is a dictionary.
        if req[key] is None, we want the subset of entries containing key.
        Otherwise we want the subset of entries such that data[k][key] == req[key]

        dfunk is a dictionary of lambdas. Each lambda is a test. If all keys are
        in the current record (another test), then it applies all lambdas to
        dfunk[x](self._data[id]). They must be True.

        key_order specifies which key is used for session_iterator2
        """
        L = list()
        for k in self.session_iterator2(key_order):
            if not key in self._data[k]:
                continue
            if req:
                if any( x not in self._data[k] for x in req ):
                    continue
                if any( req[x]!=self._data[k][x] for x in req if req[x] is not None ):
                    continue
            if dfunk:
                if any( x not in self._data[k] for x in dfunk ):
                    continue
                if any( not dfunk[x](self._data[k][x]) for x in dfunk ):
                    continue
            L.append( self._data[k][key] )

        return sorted(L) if sort_values else L


    def iter_values_for_key(self, key, req=None, dfunk=None, key_order=None, sort_values=False):
        """
        req is a dictionary.
        if req[key] is None, we want the subset of entries containing key.
        Otherwise we want the subset of entries such that data[k][key] == req[key]

        dfunk is a dictionary of lambdas. Each lambda is a test. If all keys are
        in the current record (another test), then it applies all lambdas to
        dfunk[x](self._data[id]). They must be True.

        key_order specifies which key is used for session_iterator2
        """
        return iter( self.get_values_for_key(key=key, req=req, dfunk=dfunk,
                                             key_order=key_order, sort_values=sort_values) )




    def get_known_values_for_key(self, key, sort_values=False):
        """
        """
        return set( self.get_values_for_key(key=key, sort_values=sort_values) )


    def get_all_values_for_key(self, key, subset_key=None, subset_value=None, sort_values=False):
        """
        """
        req = None if subset_key is None else {subset_key: subset_value}
        return self.get_values_for_key(key=key, req=req, dfunk=None, sort_values=sort_values)

    

